@@ -1,15 +1,9 @@
 import logging
-import sys
-from typing import Callable, Coroutine, Dict, List, Optional
+from typing import Callable, Coroutine, Dict, List, Literal, Optional, TypedDict
 from warnings import warn
 
 from dbus_fast import Variant
 
-if sys.version_info[:2] < (3, 8):
-    from typing_extensions import Literal, TypedDict
-else:
-    from typing import Literal, TypedDict
-
 from ...exc import BleakError
 from ..scanner import AdvertisementData, AdvertisementDataCallback, BaseBleakScanner
 from .advertisement_monitor import OrPatternLike
@@ -273,10 +267,7 @@ class BleakScannerBlueZDBus(BaseBleakScanner):
             advertisement_data,
         )
 
-        if self._callback is None:
-            return
-
-        self._callback(device, advertisement_data)
+        self.call_detection_callbacks(device, advertisement_data)
 
     def _handle_device_removed(self, device_path: str) -> None:
         """
