@@ -1,6 +1,12 @@
 # -*- coding: utf-8 -*-
+import sys
 
-from bleak_winrt.windows.devices.bluetooth.genericattributeprofile import GattDescriptor
+if sys.version_info >= (3, 12):
+    from winrt.windows.devices.bluetooth.genericattributeprofile import GattDescriptor
+else:
+    from bleak_winrt.windows.devices.bluetooth.genericattributeprofile import (
+        GattDescriptor,
+    )
 
 from ..descriptor import BleakGATTDescriptor
 
